import datetime
import logging
from typing import Iterator, Optional, Dict, Tuple, Any, Type, Union
from urllib.parse import urljoin, urlparse

import requests

from kada_client.models import Event, Group, Person

logger = logging.getLogger(__name__)


DEFAULT_USE_MODELS = True


class KadaEndpoints:
    """Get endpoints from an API base URL."""

    def __init__(self, url: str):
        self.base_url: str = url

    def __repr__(self):
        return "{cls.__name__}({url!r})".format(cls=type(self), url=self.base_url)

    def add_base(self, path: str) -> str:
        return urljoin(self.base_url, path)

    def list_events(self) -> str:
        return self.add_base("events/")

    def get_event(self, event_id) -> str:
        return self.add_base(f"events/{event_id}/")

    def list_persons(self) -> str:
        return self.add_base("persons/")

    def get_person(self, person_id) -> str:
        return self.add_base(f"persons/{person_id}/")

    def list_groups(self) -> str:
        return self.add_base("groups/")

    def get_group(self, group_id) -> str:
        return self.add_base(f"groups/{group_id}/")

    def list_person_events(self, person_id) -> str:
        return self.add_base(f"persons/{person_id}/events/")

    def get_person_event(self, person_id, event_id) -> str:
        return self.add_base(f"persons/{person_id}/events/{event_id}/")

    def list_event_groups(self, event_id) -> str:
        return self.add_base(f"events/{event_id}/groups/")

    def list_event_persons(self, event_id) -> str:
        return self.add_base(f"events/{event_id}/persons/")

    def get_event_person(self, event_id, person_id) -> str:
        return self.add_base(f"events/{event_id}/persons/{person_id}/")

    def list_group_events(self, group_id) -> str:
        return self.add_base(f"groups/{group_id}/events/")

    def list_group_persons(self, group_id) -> str:
        return self.add_base(f"groups/{group_id}/persons/")

    def list_person_groups(self, person_id) -> str:
        return self.add_base(f"persons/{person_id}/groups/")


class KadaClient:
    def __init__(
        self,
        url: str,
        headers: Optional[Dict[str, str]] = None,
        rewrite_url: Optional[Tuple[str, str]] = None,
        use_models: bool = DEFAULT_USE_MODELS,
        use_sessions: bool = True,
    ):
        """
        Kada API client.

        :param str base_url: Base API URL
        :param dict headers: Append extra headers to all requests
        :param tuple(str, str) rewrite_url: Rewrite URLs before sending requests (from, to)
        :param bool use_models:
            When the API returns data or a list of identifiers,
            convert to objects as defined in fs_client.models
        :param bool use_sessions: Re-use connections (default: on)
        """
        self.urls = KadaEndpoints(url)
        self.headers: dict = {} if headers is None else headers
        self.rewrite_url: Optional[Tuple[str, str]] = rewrite_url
        self.use_models: bool = use_models

        if use_sessions:
            self.session = requests.Session()
        else:
            self.session = requests

    def _build_request_headers(self, headers: Optional[dict]) -> dict:
        request_headers = {}
        for h in self.headers:
            request_headers[h] = self.headers[h]
        for h in headers or ():
            request_headers[h] = headers[h]
        return request_headers

    def call(
        self,
        method_name: str,
        url: str,
        headers: Optional[Dict[str, str]] = None,
        params: Optional[Dict[str, Any]] = None,
        return_response: bool = False,
        **kwargs,
    ):
        headers = self._build_request_headers(headers)
        if params is None:
            params = {}
        if self.rewrite_url is not None:
            url = url.replace(*self.rewrite_url)
        logger.debug(
            "Calling %s %s with params=%r", method_name, urlparse(url).path, params
        )
        r = self.session.request(
            method_name, url, headers=headers, params=params, **kwargs
        )
        if r.status_code in (500, 400, 401):
            logger.warning("Got HTTP %d: %r", r.status_code, r.content)
        if return_response:
            return r
        if r.status_code == 404:
            return None
        r.raise_for_status()
        return r.json()

    def get(self, url, **kwargs):
        return self.call("GET", url, **kwargs)

    def post(self, url, **kwargs):
        return self.call("POST", url, **kwargs)

    def put(self, url, **kwargs):
        return self.call("PUT", url, **kwargs)

    def depaginate(self, page):
        """
        Yield all items, following links from `page`.
        """
        while page:
            page.raise_for_status()
            data = page.json()
            items = data.get("results")
            for item in items:
                yield item
            next_page = data.get("next")
            if next_page:
                page = self.get(next_page, return_response=True)
            else:
                break

    def maybe_object(self, object_type: Type, obj: dict):
        if not self.use_models:
            return obj
        if obj is None:
            return None
        return object_type.from_dict(obj)

    def maybe_list_of_objects(self, cls, items):
        for item in items:
            yield self.maybe_object(cls, item)

    def list_events(self) -> Iterator[Union[Event, dict]]:
        """
        List events.
        """
        url = self.urls.list_events()
        first_page = self.get(url, return_response=True)
        events = self.depaginate(first_page)
        return self.maybe_list_of_objects(Event, events)

    def get_event(self, event_id) -> Optional[Event]:
        """
        Get an event by ID.

        :param int event_id: Event ID
        """
        url = self.urls.get_event(event_id)
        event = self.get(url)
        return self.maybe_object(Event, event)

    def list_persons(self, uid: Optional[str] = None) -> Iterator[Union[Person, dict]]:
        """
        List persons.
        """
        url = self.urls.list_persons()
        if uid:
            first_page = self.get(url, return_response=True, params={"uid": uid})
        else:
            first_page = self.get(url, return_response=True)
        persons = self.depaginate(first_page)
        return self.maybe_list_of_objects(Person, persons)

    def get_person(self, person_id) -> Optional[Person]:
        """
        Get a person by ID.

        :param int person_id: Person ID
        """
        url = self.urls.get_person(person_id)
        person = self.get(url)
        return self.maybe_object(Person, person)

    def get_person_event(self, person_id: int, event_id: int) -> Optional[Event]:
        """
        Fetch a specific event that may be associated with a person.
        """
        url = self.urls.get_person_event(person_id, event_id)
        person_event = self.get(url)
        return self.maybe_object(Event, person_event)

    def list_person_events(
        self, person_id: Optional[int] = None
    ) -> Iterator[Union[Event, dict]]:
        """
        List all events associated with a person.
        """
        url = self.urls.list_person_events(person_id)
        first_page = self.get(url, return_response=True)
        person_events = self.depaginate(first_page)
        yield from self.maybe_list_of_objects(Event, person_events)

    def get_event_person(self, person_id: int, event_id: int) -> Optional[Person]:
        """
        Fetch a specific person that may be associated with an event.
        """
        url = self.urls.get_event_person(event_id, person_id)
        event_person = self.get(url)
        return self.maybe_object(Person, event_person)

    def get_group(self, group_id: int) -> Optional[Group]:
        """Get a group by ID."""
        url = self.urls.get_group(group_id)
        group = self.get(url)
        return self.maybe_object(Group, group)

    def list_groups(
        self, params: Optional[dict] = None
    ) -> Iterator[Union[Group, dict]]:
        """
        List groups.

        Accepts a dict of query params. See the api schema of your
        KalenderData installation for the allowed parameters. Usually
        available under /schema/swagger-ui
        """
        extra_kwargs = {}
        if params:
            extra_kwargs["params"] = params
        url = self.urls.list_groups()
        first_page = self.get(url, return_response=True, **extra_kwargs)
        groups = self.depaginate(first_page)
        yield from self.maybe_list_of_objects(Group, groups)

    def list_event_groups(self, event_id: int) -> Iterator[Union[Group, dict]]:
        """
        List all groups associated with an event.
        """
        url = self.urls.list_event_groups(event_id)
        first_page = self.get(url, return_response=True)
        event_groups = self.depaginate(first_page)
        yield from self.maybe_list_of_objects(Group, event_groups)

    def list_group_events(self, group_id: int) -> Iterator[Union[Event, dict]]:
        """
        List all events associated with a group.
        """
        url = self.urls.list_group_events(group_id)
        first_page = self.get(url, return_response=True)
        group_events = self.depaginate(first_page)
        yield from self.maybe_list_of_objects(Event, group_events)

    def list_event_persons(
        self, event_id: Optional[int] = None
    ) -> Iterator[Union[Person, dict]]:
        """
        List all persons associated with an event.
        """
        url = self.urls.list_event_persons(event_id)
        first_page = self.get(url, return_response=True)
        event_persons = self.depaginate(first_page)
        yield from self.maybe_list_of_objects(Person, event_persons)

    def list_group_persons(self, group_id: int) -> Iterator[Union[Person, dict]]:
        """
        List all persons associated with a group.
        """
        url = self.urls.list_group_persons(group_id)
        first_page = self.get(url, return_response=True)
        group_persons = self.depaginate(first_page)
        yield from self.maybe_list_of_objects(Person, group_persons)


def get_client(config_dict):
    """
    Get a KadaClient from configuration.
    """
    return KadaClient(
        url=config_dict["url"],
        headers=config_dict.get("headers"),
        rewrite_url=config_dict.get("rewrite_url"),
    )
